import m114

-- Définition de « limite de f en x₀ = y₀ »
-- Lean n'a pas besoin de parenthèse dans f(x)
def continue_en (f : ℝ → ℝ) (x₀ : ℝ) : Prop :=
∀ ε > 0, ∃ δ > 0, ∀ x, |x - x₀| ≤ δ → |f x - f x₀| ≤ ε

-- Une suite u est une fonction de ℕ dans ℝ
-- Définition de « u tend vers l »
def limite_suite (u : ℕ → ℝ) (l : ℝ) : Prop :=
∀ ε > 0, ∃ N, ∀ n ≥ N, |u n - l| ≤ ε

-- Si f est continue en x₀ et la suite u tend vers x₀ alors la suite f ∘ u, qui envoie n sur f (u n),
-- tend vers f x₀
example (f u x₀) (hu : limite_suite u x₀) (hf : continue_en f x₀) :
  limite_suite (f ∘ u) (f x₀) :=
begin
  -- Notre but est de montrer que, pour tout ε strictement positif, il existe un entier
  -- N tel que, pour tout entier n supérieur à N, |f(u_n) - f(x₀)| est inférieur
  -- à ε.
  unfold limite_suite,
  -- Soit ε un réel strictement positif
  intros ε hε,
  /- specialize hf ε hε,
  rcases hf with ⟨δ, δ_pos, H⟩, -/
  -- Par hypothèse de continuité sur f, appliquée à cet ε strictement positif,
  -- il existe un réel δ strictement positif tel que :
  -- Pour tout x réel, si |x - x₀| ≤ δ alors |f(x) - f(x₀)| ≤ ε   (1).
  -- On fixe un tel δ.
  obtain ⟨δ, δ_pos, Hf⟩ : ∃ δ > 0, ∀ x, |x - x₀| ≤ δ → |f x - f x₀| ≤ ε,
    exact hf ε hε,
  -- Par l'hypothèse de limite sur u, appliquée au réel δ strictement
  -- positif que nous venons de fixer, il existe un entier N tel
  -- que :
  -- Pour tout entier n, si n ≥ N alors |u_n - x₀| ≤ δ     (2).
  -- On fixe un tel N.
  obtain ⟨N, Hu⟩ : ∃ (N : ℕ), ∀ (n : ℕ), n ≥ N → |u n - x₀| ≤ δ,
    exact hu δ δ_pos,
  -- Montrons que N convient à notre objectif.
  use N,
  -- Soit n un entier supérieur à N. Montrons que |f(u_n) - f(x₀)| ≤ ε.
  intros n hn,
  -- Vu la propriété (1), appliquée à u_n, il suffit de montrer que
  -- |u_n - x₀| ≤ δ.
  apply Hf,
  -- Cela découle immédiatement de la propriété (2) et de notre hypothèse sur n
  exact Hu n hn
end
