import m114

/-
Dans les exercices suivants on pourra utiliser les faits suivants :
  mul_assoc a b c : a * b * c = a * (b * c)
  mul_comm a b : a*b = b*a

La commande "rw mul_assoc a b c," va remplacer a*b*c par a*(b*c) dans
l'objectif courant.

Pour remplacer de la droite vers la gauche, on utilise
"rw ← mul_assoc a b c," qui va remplacer a*(b*c) par a*b*c dans
l'objectif courant.
-/

example (a b c : ℝ) : (a * b) * c = b * (a * c) :=
begin
  rw mul_comm a b,
  rw mul_assoc b a c,
end

example (a b c : ℝ) : (c * b) * a = b * (a * c) :=
begin
  sorry
end

example (a b c : ℝ) : a * (b * c) = b * (a * c) :=
begin
  sorry
end

/-
Reprendre l'exemple précédant en expérimentant ce qui se passe si on ne fournit
pas les arguments aux énoncés mul_assoc ou mul_comm.
Par exemple on pourra commencer par "rw ← mul_assoc,"
-/

example (a b c : ℝ) : a * (b * c) = b * (a * c) :=
begin
  sorry
end

/-
On peut aussi réécrire dans une hypothèse du contexte local, en utilisant
par exemple "rw mul_comm a b at hyp," pour remplacer a*b par b*a
dans l'hypothèse hyp.

Dans l'exemple suivant on utilise le fait
two_mul a : 2*a = a + a
-/

example (a b c d : ℝ) (hyp : c = d*a + b) (hyp' : b = a*d) : c = 2*a*d :=
begin
  rw hyp' at hyp,
  rw mul_comm d a at hyp,
  rw ← two_mul (a*d) at hyp,
  rw ← mul_assoc 2 a d at hyp,
  exact hyp,
end

/-
Dans l'exemple suivant, on pourra utiliser
sub_self x : x - x = 0
-/

example (a b c d : ℝ) (hyp : c = b*a - d) (hyp' : d = a*b) : c = 0 :=
begin
  sorry
end

/-
La méthode des deux exemples précédents ne ressemble pas à ce qu'on écrit sur
papier. Voici comment présenter les choses autrement.
À l'intérieur de chaque paire d'accolades, l'objectif est l'égalité avec
la formule précédente.
-/

example (a b c d : ℝ) (hyp : c = d*a + b) (hyp' : b = a*d) : c = 2*a*d :=
begin
  calc c = d*a + b : by { rw hyp }
  ... = d*a + a*d  : by { rw hyp' }
  ... = a*d + a*d  : by { rw mul_comm d a }
  ... = 2*(a*d)    : by { rw two_mul }
  ... = 2*a*d      : by { rw mul_assoc },
end

/-
Reprenons l'autre exemple par cette méthode. Le plus simple est d'écrire
d'abord tout le calcul avec ": by {}" à la fin de chaque ligne puis
de compléter les justifications.
-/

example (a b c d : ℝ) (hyp : c = b*a - d) (hyp' : d = a*b) : c = 0 :=
begin
  sorry
end

/-
Les démonstrations précédentes ont épuisé notre patience avec "mul_comm"
et "mul_assoc". Il est temps de faire travailler un peu plus l'ordinateur.
-/

example (a b c d : ℝ) (hyp : c = d*a + b) (hyp' : b = a*d) : c = 2*a*d :=
begin
  calc c = d*a + b : by { rw hyp }
  ... = d*a + a*d : by { rw hyp' }
  ... = 2*a*d : by { compute },
end

example (a b c : ℝ) : a * (b * c) = b * (a * c) :=
begin
  sorry
end

example (a b : ℝ) : (a + b) + a = 2*a + b :=
begin
  sorry
end

/-
Si vous êtes en avance et pas encore convaincus par "compute",
démontrer le lemme suivant sans compute, mais avec
pow_two x : x^2 = x*x
mul_sub a b c : a*(b-c) = a*b - a*c
add_mul a b c : (a+b)*c = a*c + b*c
add_sub a b c : a + (b - c) = (a + b) - c
sub_sub a b c : a - b - c = a - (b + c)
add_zero a : a + 0 = a
-/

example (a b : ℝ) : (a + b)*(a-b) = a^2 - b^2 :=
begin
  sorry
end
